﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace FTPClient
{
    public class FtpClient
    {
        public static Socket socket;
        public static Socket ftpSocket;
        public static IPEndPoint ipEndp;
        public static IPEndPoint ftpIpEndp;
        public static NetworkStream ns;
        public static NetworkStream ftpNs;
        public static StreamReader sr;
        public static StreamReader ftpSr;
        public static StreamWriter sw;
        public static StreamWriter ftpSw;
        public string status;

        public string username { get; set; }
        public string password { get; set; }
        public string host { get; set; }
        public int port { get; set; }
        public int ftpPort { get; set; }

        // Constructor
        public FtpClient(string username, string password, string host, int port)
        {
            this.username = username;
            this.password = password;
            this.host = host;
            this.port = port;
            this.ftpPort = -1;
        }

        // Initialize socket and connect to server
        public void Initialize()
        {
            socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPAddress[] ips = Dns.GetHostAddresses(host);
            IPAddress address;
            try
            {
                address = ips[1];
            } catch (IndexOutOfRangeException)
            {
                address = ips[0];
            }
            
            ipEndp = new IPEndPoint(address, port);
            
            try
            {
                socket.Connect(ipEndp);
            } catch (Exception e)
            {
                Console.Write(e);
                Console.ReadKey();
            }

            ns = new NetworkStream(socket);
            sr = new StreamReader(ns);
            sw = new StreamWriter(ns);
            string response = sr.ReadLine();
            handleResponse(response);
        }

        private void CreateSocket()
        {
            ftpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPAddress[] ips = Dns.GetHostAddresses(host);
            IPAddress address;
            try
            {
                address = ips[1];
            } catch (IndexOutOfRangeException)
            {
                address = ips[0];
            }
            
            ftpIpEndp = new IPEndPoint(address, ftpPort);

            try
            {
                ftpSocket.Connect(ftpIpEndp);
            }
            catch (Exception e)
            {
                Console.Write(e);
                Console.ReadKey();
            }

            ftpNs = new NetworkStream(ftpSocket);
            ftpSr = new StreamReader(ftpNs);
            ftpSw = new StreamWriter(ftpNs);
        }

        /// <summary>
        /// Handles the epsv response and on 229 status creates a socket for data 
        /// </summary>
        /// <param name="response">Response from the FTP server</param>
        private void handleEpsvResponse(string response)
        {
            status = response.Substring(0, 3);
            if (status == "229")
            {
                string newPort = response.Substring(0, response.LastIndexOf('|'));
                int lastIndex = newPort.LastIndexOf('|');
                int endIndex = newPort.Length - lastIndex - 1;
                newPort = newPort.Substring(lastIndex+1, endIndex);
                int portInt = int.Parse(newPort);

                this.ftpPort = portInt;

                CreateSocket();
            }

            handleResponse(response);
        }

        /// <summary>
        /// Handles normal response from server, if response has dash after status then
        /// ask for new line. Since it indicates multiline response.
        /// </summary>
        /// <param name="response">Response from server</param>
        private void handleResponse(string response)
        {
            status = response.Substring(0, 3);
            bool loop = true;
            string resp = String.Empty;
            Console.WriteLine(response);
            if(response[3] == '-')
            {
                while (loop)
                {
                    resp = sr.ReadLine();
                    if (resp[3] == '-')
                    {
                        Console.WriteLine(resp);
                    }
                    else
                    {
                        Console.WriteLine(resp);
                        loop = false;
                    }
                }
            }
        }

        /// <summary>
        /// Handles PASV response 
        /// TODO: Parse IP from response
        /// </summary>
        /// <param name="response">Response from server</param>
        private void handlePasvResponse(string response)
        {
            status = response.Substring(0,3);
            int start = response.IndexOf('(');
            int end = response.IndexOf(')');
            string portsString = response.Substring(start+1, end-start-1);
            string[] ports = portsString.Split(',');
            int port = int.Parse(ports[ports.Length - 2]) * 256 + int.Parse(ports[ports.Length - 1]);

            if (status == "227")
            {
                this.ftpPort = port;

                CreateSocket();
            }
        }

        /// <summary>
        /// Handles login to FTP server
        /// </summary>
        public void LogIn()
        {
            bool conversation = true;
            sw.WriteLine("USER " + username);
            sw.Flush();
            handleResponse(sr.ReadLine());
            while (conversation)
            {
                switch (status)
                {
                    case "220":
                        sw.WriteLine("USER " + username);
                        sw.Flush();
                        handleResponse(sr.ReadLine());
                        break;
                    case "230":
                        conversation = false;
                        break;
                    case "331":
                        sw.WriteLine("PASS " + password);
                        sw.Flush();
                        handleResponse(sr.ReadLine());
                        break;
                    default:
                        Console.WriteLine("Error in login");
                        conversation = false;
                        break;
                }
            }
        }

        public void Pasv()
        {
            sw.WriteLine("PASV");
            sw.Flush();
            handlePasvResponse(sr.ReadLine());
        }

        /// <summary>
        /// Starts epsv transfer
        /// </summary>
        public void Epsv()
        {
            sw.WriteLine("EPSV");
            sw.Flush();
            handleEpsvResponse(sr.ReadLine());
        }

        /// <summary>
        /// Initiates LIST command
        /// </summary>
        public void List()
        {
            sw.WriteLine("LIST");
            sw.Flush();
            handleResponse(sr.ReadLine());
            bool listen = true;
            while(listen)
            {
                switch(status)
                {
                    case "425":
                        listen = false;
                        break;
                    case "150":
                        string response = ftpSr.ReadLine();

                        if (response != null)
                        {
                            Console.WriteLine(response);
                        }
                        else { listen = false; }
                        handleResponse(sr.ReadLine());
                        break;
                    default:
                        listen = false;
                        break;

                }
            }
        }

        /// <summary>
        /// Sends the retrieve file command to server
        /// </summary>
        public void Retr(string file)
        {
            sw.WriteLine("MODE S");
            sw.Flush();
            handleResponse(sr.ReadLine());
            string message = @"RETR " + file;
            sw.WriteLine(message);
            sw.Flush();

            handleResponse(sr.ReadLine());
            byte[] buffer = new byte[2048];
            string last = string.Empty;

            // TODO: Change default to break
            // and cases just for statuses that need actions.
            string path = @"C:\Kurssit\ties323\data.txt";
            if (File.Exists(path)) { File.Delete(path); }
            using (var sw = new StreamWriter(path))
            switch(status)
            {
                case "150":
                    while (true)
                    {
                        ftpSocket.Receive(buffer);
                        string response = Encoding.UTF8.GetString(buffer);
                        if(response.Equals(last)) 
                        {
                            handleResponse(sr.ReadLine());
                            CloseData();
                            return; 
                        }
                        if (response != null)
                        {
                            sw.WriteLine(response);
                            last = response;
                        }
                        else { break; }
                    }
                        break;
                default:
                    Console.WriteLine("Error in retreiving file");
                    break;
            }
        }

        public void Quit()
        {
            sw.WriteLine("QUIT");
            sw.Flush();
            handleResponse(sr.ReadLine());
            TearDown();
        }

        private void TearDown()
        {
            CloseFtp();
            CloseData();
        }

        private void CloseFtp()
        {
            sw.Close();
            sr.Close();
            ns.Close();
            socket.Close();
        }

        private void CloseData() 
        {
            ftpSw.Close();
            ftpSr.Close();
            ftpNs.Close();
            ftpSocket.Close();
        }

        public void GetDirectory()
        {
            try
            {   
                socket.Connect(ipEndp);
            } catch (Exception e)
            {
                Console.Write(e);
                Console.ReadKey();
            }

            NetworkStream ns = new NetworkStream(socket);

            StreamReader sr = new StreamReader(ns);
            StreamWriter sw = new StreamWriter(ns);

            string viesti = sr.ReadLine();
            Console.WriteLine(viesti);
        }

    }
}