﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.IO;
using System.Collections.Generic;
using System.Threading;

namespace SMTPServer
{
    class Server
    {
        static void Main(string[] args)
        {
            List<string> inbox = new List<string>();
            IPEndPoint iep_smtp = new IPEndPoint(IPAddress.Loopback, 25000);
            IPEndPoint iep_pop = new IPEndPoint(IPAddress.Loopback, 110);
            IPEndPoint iep_imap = new IPEndPoint(IPAddress.Loopback, 143);
            
            bool listening = true;
            List<string> mails = new List<string>();

            while (listening)
            {
                HandlePOP3Client(iep_pop, inbox);
                HandleSMTPClient(iep_smtp);
                HandleIMAPClient(iep_imap);
            }
            
            Console.ReadKey();
        }

        // This is enum for keeping up with the state of the conversation
        enum POPTila : UInt16
        {
            Aloitus,
            Authorization,
            Transaction,
            Update
        }

        enum AuthState : UInt16
        {
            UserState,
            PassState
        }
        
        // Prints the port of the given socket
        public static void PrintListeningPort(Socket socket)
        {
            string port = ((IPEndPoint)socket.LocalEndPoint).Port.ToString();
            Console.WriteLine("Palvelin kuuntelee porttia: {0}", port);
        }

        public static void threadListenPop(object obj)
        {
            Socket scon = obj as Socket;
            byte[] data = new byte[2048];
            IPEndPoint sender = new IPEndPoint(IPAddress.Any, 0);
            EndPoint Remote = (EndPoint)(sender);
            List<string> inbox = new List<string> { "asd", "asdasd" }; // temporary
            try
            {
                scon.Listen(1);
                PrintListeningPort(scon);
                Socket newSocket = scon.Accept();
                string ip = ((IPEndPoint)(newSocket.RemoteEndPoint)).Address.ToString();
                string asiakas_port = ((IPEndPoint)(newSocket.RemoteEndPoint)).Port.ToString();
                Console.WriteLine("Uusi asiakas: {0} portista: {1}", ip, asiakas_port);

                POPTila tila = POPTila.Aloitus;
                AuthState authState = AuthState.UserState;

                bool keskustelu = true;

                while (keskustelu)
                {
                    byte[] buffer = new byte[2048];
                    string asiakas_vastaus = "";
                    int state;
                    switch (tila)
                    {
                        case POPTila.Aloitus:
                            VastaaAsiakas("+OK TIES323 POP3 server ready", newSocket);
                            tila = POPTila.Authorization;
                            break;
                        case POPTila.Authorization:
                            newSocket.Receive(buffer);
                            asiakas_vastaus = TrimVastaus(Encoding.UTF8.GetString(buffer), newSocket);

                            state = CheckAuthorization(asiakas_vastaus, authState, newSocket);
                            if (state == 1) { authState = AuthState.PassState; }
                            if (state == 2) { authState = AuthState.UserState; tila = POPTila.Transaction; }
                            if (state == 3) { keskustelu = false; }
                            break;
                        case POPTila.Transaction:
                            newSocket.Receive(buffer);
                            asiakas_vastaus = TrimVastaus(Encoding.UTF8.GetString(buffer), newSocket);

                            state = HandleTransaction(asiakas_vastaus, newSocket, inbox);
                            if (state == 1) { tila = POPTila.Update; }
                            break;
                        case POPTila.Update:
                            HandleQuit(newSocket);
                            keskustelu = false;
                            break;
                        default:
                            VastaaAsiakas("Some error happened... shutting down", newSocket);
                            keskustelu = false;
                            break;
                    }
                }
            }
            catch (SocketException ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        enum ImapState : UInt16
        {
            Unaunthenticated,
            Authenticated,
            Selected,
            Logout
        }

        public static void threadListenImap(object obj)
        {
            Socket scon = obj as Socket;
            IPEndPoint sender = new IPEndPoint(IPAddress.Any, 0);
            EndPoint Remote = (EndPoint)(sender);
            
            try
            {
                scon.Listen(1);
                PrintListeningPort(scon);
                Socket newSocket = scon.Accept();

                string ip = ((IPEndPoint)(newSocket.RemoteEndPoint)).Address.ToString();
                string asiakas_port = ((IPEndPoint)(newSocket.RemoteEndPoint)).Port.ToString();
                Console.WriteLine("Uusi IMAP asiakas: {0} portista: {1}", ip, asiakas_port);

                ImapState state = ImapState.Unaunthenticated;
                VastaaAsiakas("OK IMAP4rev1 Service Ready", newSocket);
                bool keskustelu = true;
                int counter = 1;
                string asiakas_vastaus = "";
                string vastaus = "";

                while(keskustelu)
                {
                    byte[] buffer = new byte[2048];
                    newSocket.Receive(buffer);
                    asiakas_vastaus = TrimVastaus(Encoding.UTF8.GetString(buffer), newSocket);
                    switch(asiakas_vastaus.Split(' ')[1])
                    {
                        case "login":
                            bool valid = CheckImapLogin(asiakas_vastaus);
                            if (valid)
                            {
                                vastaus = "a" + counter.ToString("000");
                                vastaus += " OK LOGIN completed";
                                VastaaAsiakas(vastaus, newSocket);
                                counter++;
                            } 
                            break;
                        // List is missing logic from different char combinations etc...
                        case "list":
                            HandleImapList(asiakas_vastaus, newSocket);
                            vastaus = "a" + counter.ToString("000");
                            VastaaAsiakas(vastaus + " OK LIST completed", newSocket);
                            break;
                        case "logout":
                            vastaus = "BYE IMAP4rev1 server terminating connection";
                            VastaaAsiakas(vastaus, newSocket);
                            vastaus = "a" + counter.ToString("000");
                            keskustelu = false;
                            vastaus += " OK LOGOUT completed";
                            VastaaAsiakas(vastaus, newSocket);
                            break;
                    }
                        
                }
            } catch (SocketException)
            {

            }
        }

        private static void HandleImapList(string str, Socket socket)
        {
            List<string> mail = Mailbox.getMail();
            string targetBox = str.Split(' ')[2];
            for (int i = 0; i < mail.Count; i++)
            {
                string vastaus = "* LIST (here more info on box) \"/\" " + targetBox + "/Sent";
                VastaaAsiakas(vastaus, socket);
            }
        }

        private static bool CheckImapLogin(string str)
        {
            string[] arr = str.Split(' ');
            return arr[2] == "mrc" && arr[3] == "secret";
        }

        public static void threadListenSmtp(object obj)
        {
            Socket scon = obj as Socket;
            byte[] data = new byte[2048];
            IPEndPoint sender = new IPEndPoint(IPAddress.Any, 0);
            EndPoint Remote = (EndPoint)(sender);

            try
            {
                scon.Listen(1);
                PrintListeningPort(scon);
                Socket newSocket = scon.Accept();

                string ip = ((IPEndPoint)(newSocket.RemoteEndPoint)).Address.ToString();
                string asiakas_port = ((IPEndPoint)(newSocket.RemoteEndPoint)).Port.ToString();
                Console.WriteLine("Uusi asiakas: {0} portista: {1}", ip, asiakas_port);
                // Aloitetaan keskustelu
                newSocket.Send(Encoding.UTF8.GetBytes("220 TIES323 Postipalvelin\r\n"));
                bool keskustelu = true;
                bool data_vipu = false;
                List<string> mail = new List<string>();

                while (keskustelu)
                {
                    byte[] buffer = new byte[2048];
                    newSocket.Receive(buffer);
                    string asiakas_vastaus = Encoding.UTF8.GetString(buffer);
                    Console.WriteLine("[{0}:{1}]: {2}", ip, asiakas_port, asiakas_vastaus);
                    string[] lines = asiakas_vastaus.Split(
                        new[] { "\r\n", "\r", "\n" },
                            StringSplitOptions.None
                        );

                    int tila = 0;
                    if (data_vipu)
                    {
                        Mailbox.newMail(ValidoiData(lines, newSocket));
                        //mail.Add(ValidoiData(lines, newSocket));
                        tila = 0;
                    }
                    else
                    {
                        tila = ValidoiVastaus(lines[0], newSocket, data_vipu);
                    }
                    switch (tila)
                    {
                        case 1:
                            data_vipu = true;
                            break;
                        case 2:
                            newSocket.Close();
                            keskustelu = false;
                            break;
                        default:
                            data_vipu = false;
                            break;
                    }
                }
            }
            catch (SocketException ex)
            {
                Console.WriteLine(ex.Message);
            }
        }

        private static void HandlePOP3Client(IPEndPoint iep, List<string> inbox)
        {
            try
            {
                Socket asiakas = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                asiakas.Bind(iep);
                Thread thread = new Thread(threadListenPop);
                thread.Start(asiakas);
            } catch (SocketException)
            {
            }
        }

        static void HandleQuit(Socket socket)
        {
            string vastaus = "+OK TIES323 POP3 server signing off";
            VastaaAsiakas(vastaus, socket);
        }

        static string TrimVastaus(string vastaus, Socket socket)
        {
            string ip = ((IPEndPoint)(socket.RemoteEndPoint)).Address.ToString();
            string asiakas_port = ((IPEndPoint)(socket.RemoteEndPoint)).Port.ToString();

            string[] user = vastaus.Split(
                new[] { "\r\n", "\r", "\n" },
                    StringSplitOptions.None
                );
            Console.WriteLine("[{0}:{1}]: {2}", ip, asiakas_port, user[0]);
            return user[0];
        }

        static int CheckAuthorization(string msg, AuthState state, Socket socket)
        {
            string[] array = msg.Split(' ');
            int change_state = 0;
            string new_msg = "";
            if (state == AuthState.UserState && array[0] == "USER")
            {
                if (array[1] == "joalhelk") 
                {
                    new_msg = "+OK " + array[1] + " is a real hoopy frood";
                    change_state = 1;
                } else
                {
                    new_msg = "-ERR sorry, no mailbox for " + array[1] + " here";
                }
            } else if (state == AuthState.PassState && array[0] == "PASS")
            {
                if (array[1] == "secretpass")
                {
                    new_msg = "+OK maildrop locked and ready";
                    change_state = 2;
                } else
                {
                    new_msg = "-ERR invalid password";
                }
            } else if(array[0] == "QUIT")
            {
                new_msg = "+OK TIES323 POP3 server signing off";
                change_state = 3;
            }
            VastaaAsiakas(new_msg, socket);
            return change_state;
        }

        static int HandleTransaction(string msg, Socket socket, List<string> viestit)
        {
            List<string> mail = Mailbox.getMail();
            string[] array = msg.Split(' ');
            int tila = 0;
            if (array[0] == "LIST")
            {
                int koko = 0;
                if(array.Length > 1)
                {
                    int num = Int32.Parse(array[1]);
                    if(num <= mail.Count)
                    {
                        string vastaus = string.Format("+OK {0} {1}", num, Encoding.UTF8.GetByteCount(mail[num - 1]));
                        VastaaAsiakas(vastaus, socket);
                    } else
                    {
                        string vastaus = string.Format("-ERR no such message, only {0} messages in maildrop", mail.Count);
                        VastaaAsiakas(vastaus, socket);
                    }
                } else
                {
                    for (int i = 0; i < mail.Count; i++)
                    {
                        koko += Encoding.UTF8.GetByteCount(mail[i]);
                    }
                    string vastaus = string.Format("+OK {0} messages ({1} octets)", mail.Count, koko);
                    VastaaAsiakas(vastaus, socket);

                    for(int i = 0; i < mail.Count; i++)
                    {
                        vastaus = "" + (i+1) + " " + Encoding.UTF8.GetByteCount(mail[i]);
                        VastaaAsiakas(vastaus, socket);
                    }
                    VastaaAsiakas(".", socket);
                }
            } else if (array[0] == "QUIT")
            {
                tila = 1;
            }

            return tila;
        }

        private static void HandleSMTPClient(IPEndPoint iep)
        {
            try
            {
                Socket asiakas = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                asiakas.Bind(iep);
                Thread thread = new Thread(threadListenSmtp);
                thread.Start(asiakas);
            } catch (SocketException)
            {
            }
        }

        private static void HandleIMAPClient(IPEndPoint iep)
        {
            try
            {
                Socket asiakas = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                asiakas.Bind(iep);
                Thread thread = new Thread(threadListenImap);
                thread.Start(asiakas);
            }
            catch (SocketException)
            {
            }
        }

        static int ValidoiVastaus(string vastaus, Socket socket, bool data_vipu)
        {
            int tila = 0;
            string msg = "500 Syntax error!";
            
            if(vastaus.Split(' ')[0] == "HELO")
            {
                string ip = ((IPEndPoint)(socket.RemoteEndPoint)).Address.ToString();
                msg = "250 TIES323 Postipalvelin HELO " + ip;
            } 
            else if (vastaus.Split(':')[0] == "MAIL FROM")
            {
                msg = "250 2.1.0 Sender... " + vastaus.Split(':')[1].Trim() + " you are ok.";
            } 
            else if (vastaus.Split(':')[0] == "RCPT TO")
            {
                msg = "250 2.1.5 Recipient... " + vastaus.Split(':')[1].Trim() + " is ok!";
            } 
            else if (vastaus == "DATA")
            {
                msg = "354 Enter mail, end with \".\" n a line by itself";
                tila = 1;
            }
            else if (vastaus == "QUIT")
            {
                msg = "221 2.0.0 TIES323 postipalvelin sulkee yhteyden...";
                tila = 2;
            } 

            VastaaAsiakas(msg, socket);
            return tila;
        }

        static string ValidoiData(string[] arr, Socket socket)
        {
            string[] data;
            string mail = "";

            for(int i = 0; i < arr.Length - 1; i++)
            {
                if(arr[i] == ".")
                {
                    string msg = "250 2.0.0 Message displayed on Server Screen.";
                    VastaaAsiakas(msg, socket);
                    return mail;
                }
                else
                {
                    mail = mail + arr[i] + " ";
                }
            }
            return mail;
        }

        static void VastaaAsiakas(string msg, Socket asiakas)
        {
            byte[] respond = Encoding.UTF8.GetBytes(msg + "\r\n");
            asiakas.Send(respond);
        }
    }
}
